from .insert_wrapper import InsertWrapper
from .query_wrapper import QueryWrapper
from .structures import structures
from .transaction import sql_context
from .update_wrapper import UpdateWrapper
from .wrapper import Wrapper

__all__ = ['sql_context', 'Wrapper', 'QueryWrapper', 'InsertWrapper', 'UpdateWrapper', 'structures']



import sqlite3
from sqlite3 import Connection
from typing import Any, Dict

from .executor import SqliteExecutor
from .sqlite_connection import SqliteConnection
from .table_info import TableField, TableInfo
from ..protocol import IExecutor, IDatabaseConnection


def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class SqliteDataSource:
    def __init__(self, name: str, conf: Dict[str, Any]):
        self.name = name
        self.src = conf['src']
        self.executor: IExecutor = SqliteExecutor(self)

    def get_name(self) -> str:
        return self.name

    # noinspection PyMethodMayBeStatic
    def get_type(self) -> str:
        return 'sqlite'

    def get_executor(self) -> IExecutor:
        return self.executor

    def get_connection(self) -> IDatabaseConnection:
        conn: Connection = sqlite3.connect(self.src)
        conn.row_factory = dict_factory
        sqlite_connection: SqliteConnection = SqliteConnection(conn)
        return sqlite_connection

    # noinspection PyMethodMayBeStatic
    def get_default_database(self) -> str:
        return ''

    def load_structure(self, database: str, table: str) -> any:
        conn = self.get_connection()
        c = conn.cursor()
        try:
            result = c.execute(f'PRAGMA table_info({table})')
            rows = result.fetchall()
            table_fields = []
            for row in rows:
                table_field = TableField(cid=row['cid'],
                                         name=row['name'],
                                         type_=row['type'],
                                         notnull=row['notnull'],
                                         dflt_value=row['dflt_value'],
                                         pk=row['pk'])
                table_fields.append(table_field)

            table_info = TableInfo(table=table, table_fields=table_fields)
            return table_info.model
        finally:
            c.close()
            conn.close()

    def ping(self, seconds):
        pass



from collections import OrderedDict
from typing import Any

from ..exception import UnsupportedException


class LRUCache:
    def __init__(self, capacity: int):
        self._cache = OrderedDict()
        self._capacity = capacity

    def get(self, key: str) -> Any:
        if key not in self._cache:
            return None
        else:
            self._cache.move_to_end(key)  # Mark as recently used
            return self._cache[key]

    def set(self, key: str, value: Any) -> None:
        if key in self._cache:
            self._cache.move_to_end(key)  # Mark as recently used
        self._cache[key] = value
        if len(self._cache) > self._capacity:
            self._cache.popitem(last=False)  # Remove least recently used item

    def remove(self, key: str) -> None:
        if key in self._cache:
            del self._cache[key]

    def remove_prefix(self, prefix: str) -> None:
        if prefix == '':
            raise UnsupportedException('Prefix cannot be empty')

        for key in list(self._cache.keys()):
            if key.startswith(prefix):
                del self._cache[key]



from seal.ws import dispatcher
from loguru import logger


def handler(path: str):
    def wrapper(func):
        dispatcher.register(path, func)
        logger.info(f'Registered websocket route: {path}')
        return func

    return wrapper



from typing import Protocol, Tuple, List

from ..result import Result, Results


class IExecutor(Protocol):

    def find(self, sql: str, args: Tuple[any, ...], model: object) -> Result | None:
        ...

    def find_list(self, sql: str, args: Tuple[any, ...], model: object) -> Results | None:
        ...

    def count(self, sql: str, args: Tuple[any, ...]) -> int | None:
        ...

    def update(self, sql: str, args: Tuple[any, ...]) -> int | None:
        ...

    def insert(self, sql: str, args: Tuple[any, ...]) -> int | None:
        ...

    def insert_bulk(self, sql: str, args: List[Tuple[any, ...]]) -> int | None:
        ...

    def custom_query(self, sql: str, args=Tuple[any, ...]) -> Results | None:
        ...

    def custom_update(self, sql: str, args=Tuple[any, ...]) -> int | None:
        ...



import threading
from datetime import datetime, timedelta
from typing import Any, Dict

import jwt
from loguru import logger

from seal.db.protocol import IDataSource
from seal.db.result import Results
from .cache import LRUCache, Cache
from .config import configurator
from .context import web_context, WebContext
from .db import Wrapper, sql_context, InsertWrapper, QueryWrapper, UpdateWrapper, structures
from .router import get, post, put, delete


class Seal:

    def __init__(self):
        self._initialized = False
        self.data_source_dict: Dict[str, IDataSource] = {}
        self._lru_cache: LRUCache = LRUCache(102400)
        self._cache = Cache()

        self.get = get
        self.post = post
        self.put = put
        self.delete = delete

        self._web_context = web_context

    def init(self, config_path: str, init_loguru: bool = True, init_database: bool = True):
        configurator.load(config_path)
        self._initialized = True

        if init_loguru:
            logger.add(self.get_config('seal', 'loguru', 'path'),
                       rotation=self.get_config('seal', 'loguru', 'rotation'),
                       retention=self.get_config('seal', 'loguru', 'retention'),
                       level=self.get_config('seal', 'loguru', 'level'))

        if init_database:
            data_source_config = self.get_config('seal', 'data_source')
            for data_source_name, data_source_conf in data_source_config.items():
                data_source = self.get_config('seal', 'data_source', data_source_name)
                if 'mysql' == data_source.get('dialect'):
                    from .db.mysql.mysql_data_source import MysqlDataSource
                    mysql_ds = MysqlDataSource(name=data_source_name, conf=data_source_conf)
                    self.data_source_dict[data_source_name] = mysql_ds

                    # a new thread for keeping connections active
                    threading.Thread(target=mysql_ds.ping, args=[data_source.get('ping')], daemon=True).start()

                elif 'sqlite' == data_source.get('dialect'):
                    from .db.sqlite.sqlite_data_source import SqliteDataSource
                    self.data_source_dict[data_source_name] = SqliteDataSource(name=data_source_name, conf=data_source_conf)
                else:
                    raise ValueError(f'不支持的数据源类型: {data_source.get("dialect")}')

            if 'default' not in self.data_source_dict:
                raise ValueError('unknown default data source')

        logger.info(f'init seal with config: {config_path}')
        return self

    def data_source(self, data_source_name):
        if not self._initialized:
            raise ValueError('uninitialized seal')
        return self.data_source_dict.get(data_source_name)

    def get_config(self, *keys):
        if not self._initialized:
            raise ValueError('uninitialized seal')
        return configurator.get_config(*keys)

    def get_config_default(self, *keys, default=None):
        if not self._initialized:
            raise ValueError('uninitialized seal')
        return configurator.get_conf_default(*keys, default=default)

    def query_wrapper(self, table: str, database: str | None = None, data_source: str = 'default', disable_logical_deleted=False) -> QueryWrapper:
        if database is None:
            database = self.data_source_dict[data_source].get_default_database()

        logical_deleted_field = self.get_config_default('seal', 'orm', 'logical_deleted_field')
        if disable_logical_deleted:
            logical_deleted_field = None
        return QueryWrapper(table=table,
                            database=database,
                            data_source=self.data_source_dict[data_source],
                            tenant_field=self.get_config_default('seal', 'orm', 'tenant_field'),
                            tenant_value=self.get_config_default('seal', 'orm', 'tenant_value'),
                            logical_deleted_field=logical_deleted_field,
                            logical_deleted_value_true=self.get_config_default('seal', 'orm', 'logical_deleted_value_true'),
                            logical_deleted_value_false=self.get_config_default('seal', 'orm', 'logical_deleted_value_false'), )

    def update_wrapper(self, table: str, database: str | None = None, data_source: str = 'default', disable_logical_deleted=False) -> UpdateWrapper:
        if database is None:
            database = self.data_source_dict[data_source].get_default_database()

        logical_deleted_field = self.get_config_default('seal', 'orm', 'logical_deleted_field')
        if disable_logical_deleted:
            logical_deleted_field = None
        return UpdateWrapper(table,
                             database=database,
                             data_source=self.data_source_dict[data_source],
                             tenant_field=self.get_config_default('seal', 'orm', 'tenant_field'),
                             tenant_value=self.get_config_default('seal', 'orm', 'tenant_value'),
                             updated_by_field=self.get_config_default('seal', 'orm', 'updated_by_field'),
                             updated_at_field=self.get_config_default('seal', 'orm', 'updated_at_field'),
                             logical_deleted_field=logical_deleted_field,
                             logical_deleted_value_true=self.get_config_default('seal', 'orm', 'logical_deleted_value_true'),
                             logical_deleted_value_false=self.get_config_default('seal', 'orm', 'logical_deleted_value_false'), )

    def insert_wrapper(self, table: str, database: str | None = None, data_source: str = 'default', disable_logical_deleted=False) -> InsertWrapper:
        if database is None:
            database = self.data_source_dict[data_source].get_default_database()

        logical_deleted_field = self.get_config_default('seal', 'orm', 'logical_deleted_field')
        if disable_logical_deleted:
            logical_deleted_field = None
        return InsertWrapper(table,
                             database=database,
                             data_source=self.data_source_dict[data_source],
                             tenant_field=self.get_config_default('seal', 'orm', 'tenant_field'),
                             tenant_value=self.get_config_default('seal', 'orm', 'tenant_value'),
                             logical_deleted_field=logical_deleted_field,
                             logical_deleted_value_true=self.get_config_default('seal', 'orm', 'logical_deleted_value_true'),
                             logical_deleted_value_false=self.get_config_default('seal', 'orm', 'logical_deleted_value_false'), )

    # noinspection PyMethodMayBeStatic
    def conditions_wrapper(self) -> Wrapper:
        return Wrapper()

    def custom_query(self, data_source: str, sql: str, args=tuple[Any, ...]) -> Results:
        if data_source not in self.data_source_dict:
            raise ValueError(f'unknown data source: {data_source}')
        if sql is None:
            raise ValueError('sql is required')
        return self.data_source_dict[data_source].get_executor().custom_query(sql, args)

    def custom_update(self, data_source: str, sql: str, args=tuple[Any, ...]) -> int | None:
        if data_source not in self.data_source_dict:
            raise ValueError(f'unknown data source: {data_source}')
        if sql is None:
            raise ValueError('sql is required')
        return self.data_source_dict['default'].get_executor().custom_update(sql, args)

    # noinspection PyMethodMayBeStatic
    def get_structure(self, name: str = None, data_source: str = 'default', database: str | None = None) -> Any:
        if database is None:
            database = self.data_source_dict[data_source].get_default_database()
        if name is None:
            raise ValueError('name is required')
        structure = structures.get(data_source, database, name)
        if structure is None:
            structure = self.data_source_dict[data_source].load_structure(database, name)
            structures.register(data_source, database, name, structure)
        return structure

    def lru_cache(self) -> LRUCache:
        return self._lru_cache

    def memory_cache(self) -> Cache:
        return self._cache

    # noinspection PyMethodMayBeStatic
    def generate_token(self, **payloads):
        try:
            payloads["exp"] = datetime.now() + timedelta(seconds=configurator.get_config('seal', 'authorization', 'expire'))
            token = jwt.encode(payloads, configurator.get_config('seal', 'authorization', 'jwt_key'), algorithm="HS256")
        except Exception as e:
            raise Exception(f"Token generation failed: {e}")
        return token

    def web_context(self) -> WebContext:
        return self._web_context.get()

    # noinspection PyMethodMayBeStatic
    def begin_tx(self, data_source: str = 'default'):
        sql_context.get().begin(self.data_source_dict[data_source])

    # noinspection PyMethodMayBeStatic
    def commit_tx(self):
        sql_context.get().commit()

    # noinspection PyMethodMayBeStatic
    def rollback_tx(self):
        sql_context.get().rollback()



from .data_source_protocol import IDataSource
from .database_connection_protocol import IDatabaseConnection
from .executor_protocol import IExecutor
from .entity_protocol import IEntity

__all__ = ['IExecutor', 'IDataSource', 'IDatabaseConnection', 'IEntity']



from typing import List, Callable, Any


def group(list_: List[Any], key_func: Callable, value_func: Callable = None):
    """
    Group a list of data by a key.
    """
    grouped = {}
    for item in list_:
        key = key_func(item)
        if key not in grouped:
            grouped[key] = []
        if value_func:
            grouped[key].append(value_func(item))
        else:
            grouped[key].append(item)
    return grouped


def merge(list_: List[Any], merge_func: Callable):
    """
    Merge a list of data by a key.
    """
    merged: Any = None
    for item in list_:
        # if merged is None:
        #     merged = item
        #     continue
        merged = merge_func(merged, item)
    return merged


def key_map(list_: List[Any], key_func: Callable):
    """
    Map a list of data by a key.
    """
    mapped = {}
    for item in list_:
        key = key_func(item)
        mapped[key] = item
    return mapped



from typing import Protocol

from .database_connection_protocol import IDatabaseConnection
from .executor_protocol import IExecutor


class IDataSource(Protocol):
    def get_name(self) -> str:
        ...

    def get_type(self) -> str:
        ...

    def get_executor(self) -> IExecutor:
        ...

    def load_structure(self, database: str, table: str) -> any:
        ...

    def get_connection(self) -> IDatabaseConnection:
        ...

    def get_default_database(self) -> str:
        ...

    def ping(self, seconds):
        ...



from dataclasses import fields
from .sql_builder import build_insert, build_insert_bulk
from .structures import structures
from seal.db.protocol import IDataSource


class InsertWrapper:
    def __init__(self,
                 table,
                 database=None,
                 data_source: IDataSource = None,
                 tenant_field=None,
                 tenant_value=None,
                 logical_deleted_field=None,
                 logical_deleted_value_true=None,
                 logical_deleted_value_false=None):
        self.table = table
        if database is not None and database != '':
            self.table = f'{database}.{table}'
        self.data_source = data_source

        self.param_type = structures.get(data_source=self.data_source.get_name(),
                                         database=database or data_source.get_default_database(),
                                         table=table)
        if self.param_type is None:
            self.param_type = self.data_source.load_structure(database, table)
            structures.register(data_source=self.data_source.get_name(),
                                database=database or data_source.get_default_database(),
                                table=table,
                                structure=self.param_type)

        self.tenant_field = tenant_field
        self.tenant_value = tenant_value
        self.logical_deleted_field = logical_deleted_field
        self.logical_deleted_value_true = logical_deleted_value_true
        self.logical_deleted_value_false = logical_deleted_value_false
        self.insert_fields = []

    def insert(self, data, **options):
        if data is None:
            raise ValueError('null data')

        if isinstance(data, dict):
            self.handle_data_public_fields(data, True)
        else:
            self.handle_data_public_fields(data, False)
        if 'duplicated_key_update' in options:
            sql, args = build_insert(self, data, options['duplicated_key_update'])
        else:
            sql, args = build_insert(self, data)
        return self.data_source.get_executor().insert(sql, args)

    def insert_bulk(self, data_list, **options):
        if data_list is None or len(data_list) == 0:
            raise ValueError('null data')

        if isinstance(data_list[0], dict):
            self.handle_data_list_public_fields(data_list, True)
        else:
            self.handle_data_list_public_fields(data_list, False)

        if 'duplicated_key_update' in options:
            sql, args = build_insert_bulk(self, data_list, duplicated_key_update=options['duplicated_key_update'])
        elif 'duplicated_key_ignore' in options:
            sql, args = build_insert_bulk(self, data_list, duplicated_key_ignore=options['duplicated_key_ignore'])
        else:
            sql, args = build_insert_bulk(self, data_list)
        return self.data_source.get_executor().insert_bulk(sql, args)

    # def insert_iterator(self, data_list, **options):
    #     if data_list is None or len(data_list) == 0:
    #         raise ValueError('null data')
    #
    #     if isinstance(data_list[0], dict):
    #         self.handle_data_list_public_fields(data_list, True)
    #     else:
    #         self.handle_data_list_public_fields(data_list, False)
    #
    #     data_iterator = build_insert_iterator(self, data_list, **options)
    #     return self.data_source.get_executor().insert_interator(data_iterator)

    def handle_data_public_fields(self, data, is_dict):
        if is_dict:
            if self.logical_deleted_field is not None:
                if self.logical_deleted_value_true is None or self.logical_deleted_value_false is None:
                    raise ValueError('logic delete field and value is required')
                data[self.logical_deleted_field] = self.logical_deleted_value_false
            if self.tenant_field is not None:
                if self.tenant_value is None:
                    raise ValueError('tenant value not set')
                data[self.tenant_field] = self.tenant_value

            keys = data.keys()
            self.insert_fields = [f.name for f in fields(self.param_type) if
                                  f.name != 'id' and f.name in keys]
        else:
            if self.logical_deleted_field is not None:
                if self.logical_deleted_value_true is None or self.logical_deleted_value_false is None:
                    raise ValueError('logic delete field and value is required')
                setattr(data, self.logical_deleted_field, self.logical_deleted_value_false)
            if self.tenant_field is not None:
                if self.tenant_value is None:
                    raise ValueError('tenant value not set')
                setattr(data, self.tenant_field, self.tenant_value)

            self.insert_fields = [f.name for f in fields(self.param_type) if
                                  f.name != 'id' and getattr(data, f.name) is not None]

    def handle_data_list_public_fields(self, data_list, is_dict):
        if is_dict:
            if self.logical_deleted_field is not None:
                if self.logical_deleted_value_true is None or self.logical_deleted_value_false is None:
                    raise ValueError('logic delete field and value is required')
                map(lambda x: x.update({self.logical_deleted_field: self.logical_deleted_value_false}), data_list)
            if self.tenant_field is not None:
                if self.tenant_value is None:
                    raise ValueError('tenant value not set')
                map(lambda x: x.update({self.tenant_field: self.tenant_value}), data_list)

            keys = data_list[0].keys()
            self.insert_fields = [f.name for f in fields(self.param_type) if
                                  f.name != 'id' and f.name in keys]
        else:
            if self.logical_deleted_field is not None:
                if self.logical_deleted_value_true is None or self.logical_deleted_value_false is None:
                    raise ValueError('logic delete field and value is required')
                map(lambda x: setattr(x, self.logical_deleted_field, self.logical_deleted_value_false), data_list)
            if self.tenant_field is not None:
                if self.tenant_value is None:
                    raise ValueError('tenant value not set')
                map(lambda x: setattr(x, self.tenant_field, self.tenant_value), data_list)

            self.insert_fields = [f.name for f in fields(self.param_type) if
                                  f.name != 'id' and getattr(data_list[0], f.name) is not None]



from typing import List

from loguru import logger

from ..protocol import IDatabaseConnection
from ..protocol.data_source_protocol import IDataSource
from ..result import Result, Results
from ..transaction import sql_context


class MysqlExecutor:

    def __init__(self, data_source: IDataSource):
        self.data_source = data_source

    def find(self, sql: str, args: tuple[any, ...], model: object) -> Result | None:
        self.debug(sql, args)

        connection: IDatabaseConnection = self.get_connection()
        cursor = connection.cursor()
        try:
            sql = sql.replace('?', '%s')
            result = cursor.execute(sql, args)
            if result is None:
                return Result.empty()

            row = cursor.fetchone()
            if row is None:
                return Result.empty()

            return Result(row=row, model=model)
        except Exception as e:
            raise e
        finally:
            cursor.close()
            self.close_connection(connection)

    def find_list(self, sql: str, args: tuple[any, ...], model: object) -> Results | None:
        self.debug(sql, args)

        connection: IDatabaseConnection = self.get_connection()
        cursor = connection.cursor()
        try:
            sql = sql.replace('?', '%s')
            result = cursor.execute(sql, args)
            if result is None:
                return Results.empty()

            rows = cursor.fetchall()
            if rows is None:
                return Results.empty()

            return Results(rows=rows, model=model)
        except Exception as e:
            raise e
        finally:
            cursor.close()
            self.close_connection(connection)

    def count(self, sql: str, args: tuple[any, ...]) -> int | None:
        self.debug(sql, args)

        connection: IDatabaseConnection = self.get_connection()
        cursor = connection.cursor()
        try:
            sql = sql.replace('?', '%s')
            result = cursor.execute(sql, args)
            if result is None:
                return None

            row = cursor.fetchone()
            if row is None:
                return None

            return row['COUNT(1)']
        except Exception as e:
            raise e
        finally:
            cursor.close()
            self.close_connection(connection)

    def update(self, sql: str, args: tuple[any, ...]) -> int | None:
        self.debug(sql, args)

        connection: IDatabaseConnection = self.get_connection()
        cursor = connection.cursor()
        try:
            sql = sql.replace('?', '%s')
            result = cursor.execute(sql, args)
            if result is None:
                return None
            return result
        except Exception as e:
            raise e
        finally:
            cursor.close()
            self.close_connection(connection)

    def insert(self, sql: str, args: tuple[any, ...]) -> int | None:
        self.debug(sql, args)

        connection: IDatabaseConnection = self.get_connection()
        cursor = connection.cursor()
        try:
            sql = sql.replace('?', '%s')
            result = cursor.execute(sql, args)
            if result is None:
                return None
            return cursor.lastrowid
        except Exception as e:
            raise e
        finally:
            cursor.close()
            self.close_connection(connection)

    def insert_bulk(self, sql: str, args: List[tuple[any, ...]]) -> int | None:
        logger.debug(f'#### sql: {sql}')

        connection: IDatabaseConnection = self.get_connection()
        cursor = connection.cursor()
        try:
            sql = sql.replace('?', '%s')
            row_affected = 0
            for row_args in args:
                logger.debug(f'#### args: {row_args}')
                row_affected += cursor.execute(sql, row_args) or 0
            logger.debug(f'#### row_affected: {row_affected}')
            connection.commit()
            return row_affected
        except Exception as e:
            connection.rollback()
            raise e
        finally:
            cursor.close()
            self.close_connection(connection)

    def custom_query(self, sql: str, args: tuple[any, ...]) -> Results | None:
        self.debug(sql, args)

        connection: IDatabaseConnection = self.get_connection()
        cursor = connection.cursor()
        try:
            sql = sql.replace('?', '%s')
            result = cursor.execute(sql, args)
            if result is None:
                return Results.empty()

            rows = cursor.fetchall()
            if rows is None:
                return Results.empty()

            return Results(rows=rows)
        except Exception as e:
            raise e
        finally:
            cursor.close()
            self.close_connection(connection)

    def custom_update(self, sql: str, args: tuple[any, ...]) -> int | None:
        self.debug(sql, args)

        connection: IDatabaseConnection = self.get_connection()
        cursor = connection.cursor()
        try:
            sql = sql.replace('?', '%s')
            result = cursor.execute(sql, args)
            if result is None:
                return None
            return result
        except Exception as e:
            raise e
        finally:
            cursor.close()
            self.close_connection(connection)

    def get_connection(self) -> IDatabaseConnection:
        ctx = sql_context.get()
        connection: IDatabaseConnection = ctx.tx() or self.data_source.get_connection()

        if ctx.tx() is None:
            connection.begin()

        return connection

    # noinspection PyMethodMayBeStatic
    def close_connection(self, connection: IDatabaseConnection):
        ctx = sql_context.get()
        if ctx.tx() is None:
            connection.commit()
            connection.close()

    # noinspection PyMethodMayBeStatic
    def debug(self, sql, args):
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')
        # ctx = sql_context.get()
        # if ctx.tx() is not None:
        #     logger.debug(f'#### Transaction id: {ctx.tx_id()}')



from dataclasses import make_dataclass, field


class TableField:

    def __init__(self, field_, type_, null_, key_, default_, extra):
        self.field_ = field_
        self.type_ = type_
        self.null_ = null_
        self.key_ = key_
        self.default_ = default_
        self.extra = extra


class TableInfo:

    def __init__(self, table, table_fields: list[TableField]):
        self.table = table
        self.table_fields = table_fields

        fields = [(table_field.field_, any, field(default=None)) for table_field in self.table_fields]
        self.model: object = make_dataclass(self.table, fields=fields)



from sqlite3 import Connection


class SqliteConnection:

    def __init__(self, connection: Connection):
        self.connection = connection

    def close(self):
        self.connection.close()

    def cursor(self):
        return self.connection.cursor()

    def commit(self):
        self.connection.execute('COMMIT')

    def rollback(self):
        self.connection.execute('ROLLBACK')

    def begin(self):
        self.connection.execute("BEGIN TRANSACTION")

    def ping(self, seconds):
        pass



import yaml


class Configuration:

    def __init__(self):
        self.config_dict = {}
        self._initialized = False

    def load(self, path: str):
        with open(path, 'r') as f:
            self.config_dict = yaml.load(f, Loader=yaml.FullLoader)
        self._initialized = True

    def get_config(self, *keys):
        if not self._initialized:
            raise ValueError('uninitialized')

        conf = self.config_dict
        for key in keys:
            conf = conf.get(key)
            if not conf:
                raise ValueError(f'config not found: {keys}')
        return conf

    def get_conf_default(self, *keys, default=None):
        if not self._initialized:
            raise ValueError('uninitialized')

        conf = self.config_dict
        for key in keys:
            conf = conf.get(key)
            if conf is None:
                return default
        return conf



from contextvars import ContextVar


class WebContext:
    def __init__(self, uid=None):
        self.uid = uid

    def set_uid(self, uid):
        self.uid = uid

    def get_uid(self):
        return self.uid

    def set(self, key, value):
        setattr(self, key, value)


web_context = ContextVar('web_context', default=WebContext())


from .list_utils import group, merge, key_map
from .str_utils import snake_case, camel_case
from .decorator_utils import singleton

__all__ = ['group', 'merge', 'key_map', 'snake_case', 'camel_case', 'singleton']



from typing import List

from loguru import logger

from seal.db.protocol.data_source_protocol import IDataSource
from seal.db.result import Result, Results


class SqliteExecutor:

    def __init__(self, data_source: IDataSource):
        self.data_source = data_source

    def find(self, sql: str, args: tuple[any, ...], model: object) -> Result | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return Result.empty()

            row = cursor.fetchone()
            if row is None:
                return Result.empty()

            return Result(row=row, model=model)
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def find_list(self, sql: str, args: tuple[any, ...], model: object) -> Results | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return Results.empty()

            rows = cursor.fetchall()
            if rows is None:
                return Results.empty()

            return Results(rows=rows, model=model)
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def count(self, sql: str, args: tuple[any, ...]) -> int | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return None

            row = cursor.fetchone()
            if row is None:
                return None

            return row['COUNT(1)']
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def update(self, sql: str, args: tuple[any, ...]) -> int | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            connection.commit()
            if result is None:
                return None
            return cursor.rowcount
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def insert(self, sql: str, args: tuple[any, ...]) -> int | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            connection.commit()
            if result is None:
                return None
            return cursor.lastrowid
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def insert_bulk(self, sql: str, args: List[tuple[any, ...]]) -> int | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.executemany(sql, args)
            connection.commit()
            if result is None:
                return None
            return cursor.rowcount
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    # def insert_interator(self, data_iterator):
    #     connection = self.data_source.get_connection()
    #     cursor = connection.cursor()
    #     try:
    #         data_iterator(lambda sql, args: cursor.execute(sql, args))
    #         connection.commit()
    #         return cursor.rowcount
    #     except Exception as e:
    #         logger.exception(e)
    #         raise e
    #     finally:
    #         cursor.close()
    #         connection.close()

    def custom_query(self, sql: str, args: tuple[any, ...]) -> Results | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            if result is None:
                return Results.empty()

            rows = cursor.fetchall()
            if rows is None:
                return Results.empty()

            return Results(rows=rows)
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()

    def custom_update(self, sql: str, args: tuple[any, ...]) -> int | None:
        logger.debug(f'#### sql: {sql}')
        logger.debug(f'#### args: {args}')

        connection = self.data_source.get_connection()
        cursor = connection.cursor()
        try:
            result = cursor.execute(sql, args)
            connection.commit()
            if result is None:
                return None
            return cursor.rowcount
        except Exception as e:
            logger.exception(e)
            raise e
        finally:
            cursor.close()
            connection.close()



from dataclasses import fields
from typing import List

from seal.db.protocol import IDataSource
from seal.db.result import Result, Results
from .protocol import IEntity
from .sql_builder import build_select, build_count
from .structures import structures
from .wrapper import Wrapper
from ..enum import ResultType
from ..types import Column


class QueryWrapper(Wrapper):
    def __init__(self,
                 table: str | IEntity,
                 database: str = None,
                 data_source: IDataSource = None,
                 tenant_field: Column | None = None,
                 tenant_value: any = None,
                 logical_deleted_field: Column | None = None,
                 logical_deleted_value_true: any = None,
                 logical_deleted_value_false: any = None):

        super().__init__(table=table,
                         tenant_field=tenant_field,
                         tenant_value=tenant_value,
                         logical_deleted_field=logical_deleted_field,
                         logical_deleted_value_true=logical_deleted_value_true,
                         logical_deleted_value_false=logical_deleted_value_false)

        if self.table is None or self.table == '':
            raise ValueError('table is required')

        if database is not None and database != '':
            self.table = f'{database}.{table}'

        self.data_source: IDataSource = data_source

        self.model = structures.get(data_source=self.data_source.get_name(),
                                    database=database or data_source.get_default_database(),
                                    table=table)
        if self.model is None:
            self.model = self.data_source.load_structure(database, table)
            structures.register(data_source=self.data_source.get_name(),
                                database=database or data_source.get_default_database(),
                                table=table,
                                structure=self.model)

        self.limit_ = None
        self.offset = None
        self.order_by = None
        self.field_list = []
        self.ignore_fields = []

    def select(self, *field_list) -> 'QueryWrapper':
        if self.entity is not None:
            for field in field_list:
                if not hasattr(self.entity, field):
                    raise ValueError(f'table {self.table} has no field {field}')

        self.field_list = field_list
        return self

    def ignore(self, *field_list) -> 'QueryWrapper':
        if self.entity is not None:
            for field in field_list:
                if not hasattr(self.entity, field):
                    raise ValueError(f'table {self.table} has no field {field}')

        self.ignore_fields = field_list
        return self

    def sort(self, *order_by) -> 'QueryWrapper':
        self.order_by = order_by
        return self

    def desc(self, *order_by) -> 'QueryWrapper':
        self.order_by = [f'{field} desc' for field in order_by]
        return self

    def asc(self, *order_by) -> 'QueryWrapper':
        self.order_by = [f'{field} asc' for field in order_by]
        return self

    def limit(self, limit: int) -> 'QueryWrapper':
        self.limit_ = limit
        return self

    def offset(self, offset: int) -> 'QueryWrapper':
        self.offset = offset
        return self

    def one(self, result_type: ResultType = ResultType.DICT, **options) -> dict | object:
        sql, args = self.build_statement(**options)
        result: Result | None = self.data_source.get_executor().find(sql, args, self.model)
        if result_type == ResultType.DICT:
            return result.as_dict()
        return result.get()

    def list(self, result_type: ResultType = ResultType.DICT, **options) -> List[dict] | List[object]:
        sql, args = self.build_statement(**options)
        results: Results = self.data_source.get_executor().find_list(sql, args, self.model)
        if result_type == ResultType.DICT:
            return results.as_dict()
        return results.get()

    def page(self, page: int, page_size: int, result_type: ResultType = ResultType.DICT, **options) -> (
            tuple[List[dict[str, any]], int] | tuple[List[any], int]):
        self.limit_ = page_size
        self.offset = (page - 1) * page_size
        sql, args = self.build_statement(**options)
        results: Results = self.data_source.get_executor().find_list(sql, args, self.model)
        count = self.count()
        if result_type == ResultType.DICT:
            return results.as_dict(), count
        return results.get(), count

    def d_page(self, page: int, page_size: int, **options) -> tuple[List[dict], int]:
        return self.page(page, page_size, as_dict=True, **options)

    def count(self):
        sql, args = build_count(self)
        return self.data_source.get_executor().count(sql, args)

    def build_statement(self, **options) -> tuple[str, tuple[any, ...]]:
        self.handle_public_fields(**options)

        if len(self.field_list) == 0:
            self.field_list = [field.name for field in fields(self.model) if field.name not in self.ignore_fields]
        return build_select(self)

    def build_sql(self, **options) -> str:
        sql, args = self.build_statement(**options)
        for arg in args:
            if isinstance(arg, str):
                sql = sql.replace('?', f"'{arg}'", 1)
            else:
                sql = sql.replace('?', str(arg), 1)
        return sql



class Result:
    def __init__(self, row: dict[str, any] | None = None, model: any = None):
        self.row: dict[str, any] = row
        self.model: any = model

    @staticmethod
    def empty() -> 'Result':
        return Result()

    def get(self) -> any:
        if self.row is not None:
            if self.model is None:
                raise Exception('no type specified')
            return self.model(**self.row)
        return None

    def as_dict(self) -> dict[str, any] | None:
        if self.row is not None:
            return self.row
        return None


class Results:
    def __init__(self, rows: list[dict[str, any]] = None, model: any = None):
        self.rows = rows
        self.model = model

    @staticmethod
    def empty() -> 'Results':
        return Results()

    def is_present(self) -> bool:
        return self.rows is not None and len(self.rows) > 0

    def get(self) -> list[object]:
        if self.rows is not None:
            if self.model is None:
                raise Exception('no type specified')
            return [self.model(**row) for row in self.rows]
        return []

    def as_dict(self) -> list[dict[str, any]]:
        if self.rows is not None:
            return self.rows
        return []



from typing import Protocol


class IDatabaseConnection(Protocol):
    def commit(self):
        ...

    def rollback(self):
        ...

    def begin(self):
        ...

    def cursor(self):
        ...

    def close(self):
        ...

    def ping(self, seconds):
        ...



import time


class Cache:
    def __init__(self):
        self.cache_dict = {}
        self.count = 0
        self.has_expired_key = False

    def set(self, key, value, ttl=None):
        """
        ttl unit: second
        """
        if ttl is not None:
            self.cache_dict[key] = CacheItem(key, value, expire_at=int(time.time()) + ttl)
            self.has_expired_key = True
        else:
            self.cache_dict[key] = CacheItem(key, value)

    def get(self, key):
        self.remove_expired()
        if key in self.cache_dict:
            item = self.cache_dict[key]
            if item.expire_at is not None and item.expire_at < int(time.time()):
                del self.cache_dict[key]
                return None
            return item.value
        return None

    def remove(self, key):
        if key in self.cache_dict:
            del self.cache_dict[key]

    def remove_expired(self):
        if not self.has_expired_key:
            return

        self.count += 1
        if self.count > 1000:
            self.count = 0

            will_expired_key_count = 0
            for k in list(self.cache_dict.keys()):
                item = self.cache_dict[k]
                if item.expire_at is not None:
                    will_expired_key_count += 1
                    if item.expire_at < int(time.time()):
                        del self.cache_dict[k]
            if will_expired_key_count == 0:
                self.has_expired_key = False


class CacheItem:
    def __init__(self, key, value, expire_at: int | None = None):
        self.key = key
        self.value = value
        self.expire_at = expire_at



from dataclasses import make_dataclass, field


class TableField:

    def __init__(self, cid, name, type_, notnull, dflt_value, pk):
        self.cid = cid
        self.name = name
        self.type_ = type_
        self.notnull = notnull
        self.dflt_value = dflt_value
        self.pk = pk


class TableInfo:

    def __init__(self, table, table_fields: list[TableField]):
        self.table = table
        self.table_fields = table_fields

        fields = [(table_field.name, any, field(default=None)) for table_field in self.table_fields]
        self.model: object = make_dataclass(self.table, fields)



from enum import Enum


class Operator(Enum):
    EQ = '='
    NE = '!='
    GT = '>'
    LT = '<'
    GE = '>='
    LE = '<='
    LIKE = 'LIKE'
    NOT_LIKE = 'NOT LIKE'
    IN = 'IN'
    NOT_IN = 'NOT IN'
    IS_NULL = 'IS NULL'
    IS_NOT_NULL = 'IS NOT NULL'
    BETWEEN = 'BETWEEN'
    NOT_BETWEEN = 'NOT BETWEEN'
    AND = 'AND'
    OR = 'OR'
    NOT = 'NOT'
    ADD = '+'
    SUB = '-'
    MUL = '*'
    DIV = '/'
    MOD = '%'



import datetime
from dataclasses import fields

from . import structures
from .protocol import IDataSource, IEntity
from .sql_builder import build_update, build_delete
from .wrapper import Wrapper
from ..types import Column


class UpdateWrapper(Wrapper):

    def __init__(self,
                 table: str | IEntity,
                 database: None | str = None,
                 data_source: IDataSource = None,
                 tenant_field: Column | None = None,
                 tenant_value: any = None,
                 updated_by_field: Column | None = None,
                 updated_at_field: any = None,
                 logical_deleted_field: Column | None = None,
                 logical_deleted_value_true: any = None,
                 logical_deleted_value_false: any = None):
        super().__init__(table=table,
                         tenant_field=tenant_field,
                         tenant_value=tenant_value,
                         logical_deleted_field=logical_deleted_field,
                         logical_deleted_value_true=logical_deleted_value_true,
                         logical_deleted_value_false=logical_deleted_value_false)

        if self.table is None or self.table == '':
            raise ValueError('table is required')

        if database is not None and database != '':
            self.table = f'{database}.{table}'

        self.data_source: IDataSource | None = data_source
        self.updated_by_field: any = updated_by_field
        self.updated_at_field: any = updated_at_field
        self.update_fields: dict = {}

        self.model = structures.get(data_source=self.data_source.get_name(),
                                    database=database or data_source.get_default_database(),
                                    table=table)
        if self.model is None:
            self.model = self.data_source.load_structure(database, table)
            structures.register(data_source=self.data_source.get_name(),
                                database=database or data_source.get_default_database(),
                                table=table,
                                structure=self.model)

    def set(self, field: Column, value: any) -> 'UpdateWrapper':
        self.update_fields[field] = value
        return self

    def set_all(self, entity: any) -> 'UpdateWrapper':
        if isinstance(entity, dict):
            for field in fields(self.model):
                if field.name != 'id' and field.name in entity:
                    self.update_fields[field.name] = entity[field.name]
            return self

        for field in fields(entity):
            if field.name != 'id':
                continue
            self.update_fields[field.name] = getattr(entity, field.name)
        return self

    def update(self, **options):
        if len(self.condition_tree.conditions) == 0:
            raise ValueError('unsupported update all')

        self.handle_public_fields(**options)
        sql, args = build_update(self)
        return self.data_source.get_executor().update(sql, args)

    def delete(self, **options):
        if len(self.condition_tree.conditions) == 0:
            raise ValueError('unsupported delete all')

        self.handle_public_fields(**options)

        if options.get('logical_delete', False):
            self.set(options['logical_delete'], 1)
            return self.update(**options)
        else:
            sql, args = build_delete(self)
            return self.data_source.get_executor().update(sql, args)

    def handle_update_fields(self, **options) -> 'UpdateWrapper':
        if self.updated_at_field is not None:
            self.set(self.updated_at_field, datetime.datetime.now())
        if self.updated_by_field is not None:
            self.set(self.updated_by_field, options.get('updated_by', 'system'))
        return self



class WSDispatcher:

    def __init__(self):
        self.registry = {}

    def dispatch(self, path):
        return self.registry.get(path)

    def register(self, path, func):
        self.registry[path] = func


dispatcher = WSDispatcher()



from functools import wraps


def singleton(orig_cls):
    orig_new = orig_cls.__new__
    orig_init = orig_cls.__init__
    instance = None

    def singleton_init(self, *args, **kwargs):
        pass

    @wraps(orig_cls.__new__)
    def __new__(cls, *args, **kwargs):
        nonlocal instance
        if instance is None:
            instance = orig_new(cls, *args, **kwargs)
            orig_init(instance, *args, **kwargs)

            if not hasattr(cls, 'singleton_init'):
                cls.singleton_init = singleton_init
        return instance

    @wraps(orig_cls.__init__)
    def __init__(obj, *args, **kwargs):
        orig_cls.singleton_init(obj, *args, **kwargs)

    orig_cls.__new__ = __new__
    orig_cls.__init__ = __init__
    return orig_cls



from typing import Dict, Any

from .executor import MysqlExecutor
from .mysql_connection import ConnectionPool
from .table_info import TableField, TableInfo
from ..protocol import IExecutor, IEntity


class MysqlDataSource:
    def __init__(self, name: str, conf: Dict[str, Any]):
        self.name = name
        self.default_database = conf.get('database')
        self.connection_pool = ConnectionPool(conf)
        self.executor: IExecutor = MysqlExecutor(self)

    def get_name(self) -> str:
        return self.name

    # noinspection PyMethodMayBeStatic
    def get_type(self) -> str:
        return 'mysql'

    def get_connection(self):
        return self.connection_pool.get_connection()

    def get_executor(self) -> IExecutor:
        return self.executor

    def get_default_database(self) -> str:
        return self.default_database

    def load_structure(self, database: str, table: str) -> any:
        conn = self.get_connection()
        conn.begin()
        c = conn.cursor()
        try:
            c.execute(f'show columns from {database}.{table}')
            rows = c.fetchall()
            table_fields = []
            for row in rows:
                table_field = TableField(field_=row['Field'],
                                         type_=row['Type'],
                                         null_=row['Null'],
                                         key_=row['Key'],
                                         default_=row['Default'],
                                         extra=row['Extra'])
                table_fields.append(table_field)

            table_info = TableInfo(table=table, table_fields=table_fields)
            return table_info.model
        finally:
            c.close()
            conn.commit()
            conn.close()

    def ping(self, seconds):
        self.connection_pool.ping(seconds)



import threading
import asyncio
from .ws_dispatcher import dispatcher
from websockets.server import serve
from websockets.exceptions import ConnectionClosed
from .. import seal
from loguru import logger


async def dispatch(websocket):
    path = websocket.path
    try:
        async for message in websocket:
            await dispatcher.dispatch(path)(message, websocket)
    except ConnectionClosed:
        await dispatcher.dispatch(path)(None, websocket, close=True)


async def start_server():
    async with serve(dispatch, seal.get_config('seal', 'ws', 'host'), seal.get_config('seal', 'ws', 'port')):
        logger.info(f'启动 websocket 服务, 监听 {seal.get_config("seal", "ws", "host")}:{seal.get_config("seal", "ws", "port")}')
        await asyncio.get_running_loop().create_future()  # run forever


def start_background():
    threading.Thread(target=asyncio.run, args=(start_server(),)).start()



class Structures:
    def __init__(self):
        self.structure_dict: dict[str, dict[str, any]] = {}

    def register(self, data_source: str, database: str | None, table: str, structure: any):
        if database is None:
            database = ''
        if f'{data_source}.{database}' not in self.structure_dict:
            self.structure_dict[f'{data_source}.{database}'] = {}
        self.structure_dict[f'{data_source}.{database}'][table] = structure

    def get(self, data_source: str, database: str | None, table: str) -> any:
        if database is None:
            database = ''
        if f'{data_source}.{database}' not in self.structure_dict:
            return None
        if table not in self.structure_dict[f'{data_source}.{database}']:
            return None
        return self.structure_dict[f'{data_source}.{database}'][table]


structures = Structures()



import schedule
from schedule import Job
import time
from loguru import logger


class JobManager:

    def __init__(self):
        self._jobs = {}
        self.started = False

    def register(self, name: str, job: Job, func, success_func=None, error_func=None):
        if self.started:
            raise Exception('job manager already started')

        def func_wrapper():
            try:
                success = func()
                if not success:
                    logger.error(f'job {name} failed')
                    if error_func:
                        try:
                            error_func()
                        except Exception as e:
                            logger.error(f'job {name} error_func error: {e}')
                            logger.exception(e)
                else:
                    logger.info(f'job {name} success')
                    if success_func:
                        try:
                            success_func()
                        except Exception as e:
                            logger.error(f'job {name} success_func error: {e}')
                            logger.exception(e)
            except Exception as e:
                logger.error(f'job {name} error: {e}')
                logger.exception(e)

        self._jobs[name] = {}
        self._jobs[name]['job'] = job
        self._jobs[name]['func'] = func_wrapper

        job.do(func_wrapper)

    def start(self):
        self.started = True
        while True:
            schedule.run_pending()
            time.sleep(1)



import time
from threading import Lock

import pymysql
from loguru import logger
from pymysql.connections import Connection as PyMySQLConnection
from pymysql.cursors import DictCursor

from seal.enum.connection_status import ConnectionStatus


class MysqlConnection:
    def __init__(self, connection: PyMySQLConnection, pool: 'ConnectionPool', create_time=time.time()):
        self.connection: PyMySQLConnection = connection
        self.pool: ConnectionPool = pool
        self.status: ConnectionStatus = ConnectionStatus.IDLE
        self.create_time = create_time
        self.conn_lock = Lock()

    def close(self):
        self.pool.release(self)
        self.conn_lock.release()

    def cursor(self):
        return self.connection.cursor()

    def commit(self):
        self.connection.commit()

    def rollback(self):
        self.connection.rollback()

    def begin(self):
        self.conn_lock.acquire_lock()
        self.connection.begin()

    def ping(self):
        try:
            if not self.conn_lock.acquire_lock(timeout=10):
                raise Exception(f'Connection [{id(self)}] is locked')
            self.connection.ping(reconnect=True)
        finally:
            self.conn_lock.release()

    def insert_id(self):
        return self.connection.insert_id()


class ConnectionPool:
    def __init__(self, conf):
        self.conf = conf

        self._connections: list[MysqlConnection] = []
        self._min_connections = conf['pool']['min_connections']
        self._max_connections = conf['pool']['max_connections']

        for _ in range(self._min_connections):
            mysql_connection: PyMySQLConnection = pymysql.connect(host=conf['host'],
                                                                  port=conf['port'],
                                                                  user=conf['user'],
                                                                  password=conf['password'].encode('utf-8'),
                                                                  database=conf['database'],
                                                                  cursorclass=DictCursor)
            self._connections.append(MysqlConnection(mysql_connection, self))

    def new_connection(self):
        return pymysql.connect(host=self.conf['host'],
                               port=self.conf['port'],
                               user=self.conf['user'],
                               password=self.conf['password'].encode('utf-8'),
                               database=self.conf['database'],
                               cursorclass=DictCursor)

    def get_connection(self, timeout=3):
        start_time = time.time()

        connection = self.occupy()
        if connection is not None:
            return connection

        if len(self._connections) < self._max_connections:
            mysql_connection = self.new_connection()
            connection = MysqlConnection(mysql_connection, self)
            connection.status = ConnectionStatus.OCCUPIED
            self._connections.append(connection)
            return connection

        if timeout is not None:
            raise Exception('No available connection')

        while True:
            connection = self.occupy()
            if connection is not None:
                return connection
            else:
                time.sleep(0.1)

                if timeout is not None and time.time() - start_time > timeout:
                    raise Exception('No available connection')

    def occupy(self) -> MysqlConnection | None:
        for connection in self._connections:
            if connection.status == ConnectionStatus.IDLE:
                connection.status = ConnectionStatus.OCCUPIED
                return connection
        return None

    def release(self, mysql_connection: MysqlConnection):
        if len(self._connections) > self._min_connections:
            mysql_connection.connection.close()
            self._connections.remove(mysql_connection)
        else:
            mysql_connection.status = ConnectionStatus.IDLE

    def ping(self, seconds=60):
        while True:
            time.sleep(seconds)
            ids = []
            for connection in self._connections:
                ids.append(id(connection))
                connection.ping()
            logger.info(f'active connections {ids}')



import asyncio
import re
import time
from functools import wraps

import jwt
from fastapi import FastAPI, Request, Response, HTTPException, Depends
from loguru import logger
from starlette.middleware.cors import CORSMiddleware

from ..config import configurator
from ..context import web_context
from ..exception import BusinessException
from ..model import Response as ResponseModel


async def verify_token(request: Request = Request):
    if request.method == 'OPTIONS':
        return None

    ignore = False
    path = request.url.path
    urls = configurator.get_config('seal', 'authorization', 'excludes')
    for url in urls:
        # * -> [a-zA-Z0-9_\-]*, ** -> .*
        url = url.replace('*', '[a-zA-Z0-9_\-]*').replace('**', '.*')
        if re.match(url, path):
            ignore = True
            break

    try:
        token = request.headers.get('Authorization', None)
        if token is None or token == '':
            if ignore:
                return None
            raise HTTPException(status_code=401, detail="Token is required")

        payload = jwt.decode(token,
                             configurator.get_config('seal', 'authorization', 'jwt_key'),
                             algorithms=["HS256"])
        web_context.get().set_uid(payload['uid'])
    except Exception as e:
        if ignore:
            return None
        raise HTTPException(status_code=401, detail=f"Token is invalid: {e}")


app = FastAPI(dependencies=[Depends(verify_token)])


@app.middleware("http")
async def calc_time(request: Request, call_next):
    start_time = time.time()
    response = await call_next(request)
    process_time = time.time() - start_time
    logger.info(f'{request.method} {request.url.path} {response.status_code} {process_time}')
    response.headers["X-Process-Time"] = str(process_time)
    return response


app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
    expose_headers=["*"]
)


@app.exception_handler(HTTPException)
async def http_exception_handler(request: Request, exc: HTTPException):
    return Response(status_code=exc.status_code, content=exc.detail)


def response_body(func, **kwargs):
    @wraps(func)
    async def wrapper(*fun_args, **fun_kwargs):
        try:
            task = asyncio.create_task(func(*fun_args, **fun_kwargs))
            result = await task
            if 'response_model' in kwargs and type(kwargs.get('response_model')) == type(ResponseModel):
                return ResponseModel.build(result).success()
            return result
        except BusinessException as e:
            logger.exception(e)
            return ResponseModel.build().error(message=e.message, code=e.code)
        except Exception as e:
            logger.exception(e)
            return ResponseModel.build().error(message=str(e))

    return wrapper


def get(path: str, **kwargs):
    def decorator(func):
        if 'response_model' not in kwargs and 'response_class' not in kwargs:
            kwargs['response_model'] = ResponseModel
        return app.get(path, **kwargs)(response_body(func, **kwargs))

    return decorator


def post(path: str, **kwargs):
    def decorator(func):
        if 'response_model' not in kwargs and 'response_class' not in kwargs:
            kwargs['response_model'] = ResponseModel
        return app.post(path, **kwargs)(response_body(func, **kwargs))

    return decorator


def delete(path: str, **kwargs):
    def decorator(func):
        if 'response_model' not in kwargs and 'response_class' not in kwargs:
            kwargs['response_model'] = ResponseModel
        return app.delete(path, **kwargs)(response_body(func, **kwargs))

    return decorator


def put(path: str, **kwargs):
    def decorator(func):
        if 'response_model' not in kwargs and 'response_class' not in kwargs:
            kwargs['response_model'] = ResponseModel
        return app.put(path, **kwargs)(response_body(func, **kwargs))

    return decorator



import setuptools

setuptools.setup(
    name='py-seal',  # 应用名
    version='0.10.0',  # 版本号
    description="一个python快速web开发框架",  # 描述
    author="melon",  # 作者
    packages=setuptools.find_packages(),  # 包括在安装包内的 Python 包 ，尽量和name一致及模块名
)



from typing import TypeVar
from pydantic import BaseModel

T = TypeVar('T')


class Response(BaseModel):
    code: int = 0
    message: str = ''
    data: T | None = None

    @staticmethod
    def build(data: T | None = None):
        response = Response()
        response.data = data
        return response

    def success(self, message='success'):
        self.code: int = 0
        self.message = message
        return self

    def error(self, code: int = -1, message='error'):
        self.code = code
        self.message = message
        return self



from .condition import ConditionTree, Condition
from .protocol import IEntity
from ..enum.operator import Operator
from ..types import Column


class Wrapper:
    def __init__(self,
                 table: str | IEntity | None = None,
                 tenant_field: Column | None = None,
                 tenant_value: any = None,
                 logical_deleted_field: Column | None = None,
                 logical_deleted_value_true: any = None,
                 logical_deleted_value_false: any = None):

        if isinstance(table, IEntity):
            self.table = table.__table_name__
            self.entity = table
        else:
            self.table = table
            self.entity = None

        self.condition_tree = ConditionTree()
        self.tenant_field: Column | None = tenant_field
        self.tenant_value: any = tenant_value
        self.logical_deleted_field: Column | None = logical_deleted_field
        self.logical_deleted_value_true: any = logical_deleted_value_true
        self.logical_deleted_value_false: any = logical_deleted_value_false

    def check_field(self, field: Column):
        if self.entity is not None and not hasattr(self.entity, field):
            raise ValueError(f'entity {self.entity} has no field {field}')

    def eq(self, field: Column, value: any) -> 'Wrapper':
        self.check_field(field)
        self.condition_tree.add_condition(Condition(field, value))
        return self

    def ne(self, field: Column, value: any) -> 'Wrapper':
        self.check_field(field)
        self.condition_tree.add_condition(Condition(field, value, Operator.NE))
        return self

    def gt(self, field: Column, value: any) -> 'Wrapper':
        self.check_field(field)
        self.condition_tree.add_condition(Condition(field, value, Operator.GT))
        return self

    def ge(self, field: Column, value: any) -> 'Wrapper':
        self.check_field(field)
        self.condition_tree.add_condition(Condition(field, value, Operator.GE))
        return self

    def lt(self, field: Column, value: any) -> 'Wrapper':
        self.check_field(field)
        self.condition_tree.add_condition(Condition(field, value, Operator.LT))
        return self

    def le(self, field: Column, value: any) -> 'Wrapper':
        self.check_field(field)
        self.condition_tree.add_condition(Condition(field, value, Operator.LE))
        return self

    def in_(self, field: Column, value: any) -> 'Wrapper':
        self.check_field(field)
        self.condition_tree.add_condition(Condition(field, value, Operator.IN))
        return self

    def l_like(self, field: Column, value: any) -> 'Wrapper':
        self.check_field(field)
        self.condition_tree.add_condition(Condition(field, f'%{value}', Operator.LIKE))
        return self

    def r_like(self, field: Column, value: any) -> 'Wrapper':
        self.check_field(field)
        self.condition_tree.add_condition(Condition(field, f'{value}%', Operator.LIKE))
        return self

    def like(self, field: Column, value: any) -> 'Wrapper':
        self.check_field(field)
        self.condition_tree.add_condition(Condition(field, f'%{value}%', Operator.LIKE))
        return self

    def or_(self, wrapper: 'Wrapper') -> 'Wrapper':
        wrapper.condition_tree.or_()
        self.condition_tree.add_tree(wrapper.condition_tree)
        return self

    def handle_public_fields(self, **options):
        if 'logical_deleted' in options:
            self.eq(options['logical_deleted'], 0)
            return
        if self.tenant_field is not None:
            if self.tenant_value is None:
                raise ValueError('tenant value not set')
            self.eq(self.tenant_field, self.tenant_value)

        if self.logical_deleted_field is not None:
            if self.logical_deleted_value_true is None or self.logical_deleted_value_false is None:
                raise ValueError('logical deleted field and value not set')
            self.eq(self.logical_deleted_field, self.logical_deleted_value_false)
